import sys
import argparse
try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata


def parse_args(args):

    parser = argparse.ArgumentParser()

    parser.add_argument('-b', '--backup', action='append',
                        help='<Required> List of Devices to Backup', required=True)

    # Optional argument which requires a parameter (eg. -d test)
    parser.add_argument("-n", "--name", default='sebs',
                        help='<Optional> specify a your app name.')

    # Optional verbosity counter (eg. -v, -vv, -vvv, etc.)
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=1,
        help="Verbosity (-v, -vv, etc)")

    # Specify output of "--version"
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s (version {version})".format(version=metadata.version('sebs')))

    if len(args) == 0:
        parser.print_help(sys.stderr)
        sys.exit(1)

    parsed_args = parser.parse_args(args)

    parsed_args.name = parsed_args.name if 'sebs' in parsed_args.name else f'{parsed_args.name}-sebs'

    return parsed_args



#!/usr/bin/env python3
"""
Module Docstring
"""

__authors__ = ["Levi Blaney", "Neha Singh"]
__license__ = "GPLv3"

import sys
import logging
from sebs.ec2 import Instance

log = logging.getLogger('sebs')


def main(args):

    log.info(f'Starting...')
    # Get a handler for the current EC2 instance
    server = Instance(args.name)

    # Add the requested Stateful Devices to the server
    for device in args.backup:
        server.add_stateful_device(device)

    # Make sure the Stateful Volumes are attached to this server
    server.attach_stateful_volumes()

    # Tag the Stateful Volumes so they can be found on next boot
    server.tag_stateful_volumes()

    for sv in server.backup:
        log.info(f"{sv.device_name} is {'Ready' if sv.ready else 'not Ready'}")

    log.info('Finished')
    sys.exit()



import sys
import boto3
import logging
import requests
from ec2_metadata import ec2_metadata

log = logging.getLogger('sebs')

for name in logging.Logger.manager.loggerDict.keys():
    if ('boto' in name) or ('urllib3' in name):
        logging.getLogger(name).setLevel(logging.WARNING)


class Instance:
    def __init__(self, volume_tag):
        # Create a session so we don't have to keep getting creds.
        self.session = None
        self.instance = self.get_instance()
        self.volume_tag = volume_tag
        self.backup = []

    def get_instance(self):
        log.info('Getting EC2 instance metadata.')
        try:
            instance_id = ec2_metadata.instance_id
            log.info(f'Running on {instance_id}')
        except requests.exceptions.ConnectTimeout:
            log.error(
                'Failled to get instance metadata, are you sure you are running on an EC2 instance?')
            sys.exit(1)
        except:
            t, v, _tb = sys.exc_info()

            sys.exit(1)

        try:
            self.session = boto3.session.Session(
                region_name=ec2_metadata.region)
            ec2 = self.session.resource('ec2')
            instance = ec2.Instance(instance_id)
            # We have to call load to see if we are really connected
            instance.load()
        except:
            t, v, _tb = sys.exc_info()
            log.error(f'Unexpected Error {t}: {v}')
            sys.exit(2)

        return instance

    def add_stateful_device(self, device_name):
        log.info(f'Handling {device_name}')
        sv = StatefulVolume(self.session, self.instance.id,
                            device_name, self.volume_tag)

        sv.get_status()

        self.backup.append(sv)

    def tag_stateful_volumes(self):
        log.info(f'Tagging Volumes with control tag: {self.volume_tag}')
        for sv in self.backup:
            if sv.status not in ['Duplicate', 'Missing']:
                sv.tag_volume()

    def attach_stateful_volumes(self):
        log.info(f'Attaching Volumes to {self.instance.id}')
        for sv in self.backup:
            if sv.status == 'Not Attached':
                sv.copy(ec2_metadata.availability_zone)
                sv.attach()


class StatefulVolume:
    def __init__(self, session, instance_id, device_name, tag_name):
        # Use the existing session so we dont have to keep fetching creds
        self.session = session
        self.instance_id = instance_id
        self.device_name = device_name
        self.ready = False
        self.status = 'Unknown'
        self.volume = None
        self.tag_name = tag_name
        self.ec2_client = self.session.client('ec2')
        self.ec2_resource = self.session.resource('ec2')

    def get_status(self):
        log.info(f'Checking for previous volume of {self.device_name}')
        response = self.ec2_client.describe_volumes(
            Filters=[
                {
                    'Name': f'tag:{self.tag_name}',
                    'Values': [
                        self.device_name,
                    ]
                },
            ]
        )

        log.debug(f'Response of tag search: {response}')

        if not response['Volumes']:
            log.info(f'Did not find a previous volume for {self.device_name}')
            self.status = 'New'
            log.info(
                f'Checking if {self.device_name} is mounted to {self.instance_id}.')

            response = self.ec2_client.describe_volumes(
                Filters=[
                    {
                        'Name': 'attachment.instance-id',
                        'Values': [
                            self.instance_id,
                        ]
                    },
                    {
                        'Name': 'attachment.device',
                        'Values': [
                            self.device_name,
                        ]
                    }
                ]
            )

            log.debug(f'Reponse of local attachment search: {response}')

            if not response['Volumes']:
                log.error(
                    f"Could not find EBS volume mounted at {self.device_name} for {self.instance_id}")

                self.status = 'Missing'
                return self.status

            volumeId = response['Volumes'][0]['VolumeId']
            log.info(f'No pre-existing volume for {self.device_name}')

            self.status = 'Attached'
            self.volume = self.ec2_resource.Volume(volumeId)
            log.info(f'Current volume is {volumeId} and is {self.status}')
            self.tag_volume()

        elif len(response['Volumes']) != 1:
            vol1 = response['Volumes'][0]['VolumeId']
            vol2 = response['Volumes'][1]['VolumeId']
            log.error(
                f"Found duplicate EBS volumes with tag {self.tag_name}: {vol1} and {vol2}")

            self.status = 'Duplicate'
        else:
            volumeId = response['Volumes'][0]['VolumeId']

            log.info(
                f'Found existing Volume {volumeId} for {self.device_name}')

            self.status = 'Not Attached'
            self.volume = self.ec2_resource.Volume(volumeId)

            for attachment in self.volume.attachments:
                if attachment['InstanceId'] == self.instance_id:
                    self.status = 'Attached'

        return self.status

    def tag_volume(self):
        log.info(
            f'Tagging {self.volume.volume_id} with control tag {self.tag_name}.')

        self.volume.create_tags(Tags=[
            {
                'Key': self.tag_name,
                'Value': self.device_name
            },
        ])

        self.ready = True

    def copy(self, target_az):
        if self.status != 'Not Attached':
            return self.status
        # If the current volume az is in the target AZ do nothing
        if target_az == self.volume.availability_zone:
            return self.status

        log.info(f'Copying {self.volume.volume_id} to {target_az}')

        snapshot = self.volume.create_snapshot(
            Description='Intermediate snapshot for SEBS.',
            TagSpecifications=[
                {
                    'ResourceType': 'snapshot',
                    'Tags': [
                        {
                            'Key': self.tag_name,
                            'Value': self.device_name
                        },
                    ]
                },
            ]
        )

        log.debug(f'Snapshot: {snapshot.snapshot_id}')

        # Not sure but we probably have to wait until its completed
        snapshot.wait_until_completed()

        # If we fail to create the volume we need to remove this temp snapshot

        response = self.ec2_client.create_volume(
            AvailabilityZone=target_az,
            SnapshotId=snapshot.snapshot_id,
            VolumeType='' if not self.volume.volume_type else self.volume.volume_type,
            TagSpecifications=[
                {
                    'ResourceType': 'volume',
                    'Tags': [
                        {
                            'Key': self.tag_name,
                            'Value': self.device_name,
                        },
                    ]
                },
            ]
        )

        log.debug(f'New Volume: {response}')

        # This should be the existing volume thats in the wrong AZ
        prev_volume = self.volume

        self.volume = self.ec2_resource.Volume(response['VolumeId'])

        log.info(f'Waiting on volume {self.volume.volume_id} to be avaliable.')

        waiter = self.ec2_client.get_waiter('volume_available')

        waiter.wait(VolumeIds=[self.volume.volume_id])

        # Cleanup this temporary resources
        prev_volume.delete()
        snapshot.delete()

        return self.status

    def attach(self):
        if self.status != 'Not Attached':
            return self.status

        log.info(f'Attaching {self.volume.volume_id} to {self.instance_id}')

        # Need to find and delete any current volumes
        response = self.ec2_client.describe_volumes(
            Filters=[
                {
                    'Name': 'attachment.instance-id',
                    'Values': [
                        self.instance_id,
                    ]
                },
                {
                    'Name': 'attachment.device',
                    'Values': [
                        self.device_name,
                    ]
                },
            ]
        )

        log.debug(f'Existing Volume: {response}')

        if response['Volumes']:
            prev_volume = self.ec2_resource.Volume(
                response['Volumes'][0]['VolumeId'])

            log.info(
                f'Detaching curent Volume {prev_volume.volume_id} attached to {self.instance_id}')

            prev_volume.detach_from_instance(
                Device=self.device_name,
                InstanceId=self.instance_id
            )

            waiter = self.ec2_client.get_waiter('volume_available')

            waiter.wait(VolumeIds=[prev_volume.volume_id])

            log.info('Waiting on detachment and then deleting.')

            prev_volume.delete()

        log.info(
            f'Attaching sebs {self.volume.volume_id} to {self.instance_id}')

        self.volume.attach_to_instance(
            Device=self.device_name,
            InstanceId=self.instance_id
        )

        waiter = self.ec2_client.get_waiter('volume_in_use')

        waiter.wait(VolumeIds=[self.volume.volume_id])

        self.status = 'Attached'

        return self.status



import unittest
from io import StringIO
from unittest.mock import patch
from sebs.cli import parse_args

try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata


class TestArugmentParsing(unittest.TestCase):

    def test_version(self):

        expected_version = metadata.version('sebs')
        with patch('sys.stdout', new=StringIO()) as fakeOutput:
            try:
                parse_args(['--version'])
            except:
                pass

        output = fakeOutput.getvalue().strip()

        self.assertTrue(expected_version in output,
                        'Should display the proper version.')

    def test_single_backup(self):
        args = parse_args(['-b', 'test'])
        self.assertEqual(len(args.backup), 1)

    def test_multiple_backup(self):
        args = parse_args(['-b', 'test1', '-b', 'test2'])

        self.assertEqual(len(args.backup), 2)

    def test_default_name(self):
        args = parse_args(['-b', 'test1', '-b', 'test2'])

        self.assertEqual(args.name, 'sebs')

    def test_override_name(self):
        args = parse_args(
            ['-b', 'test1', '-b', 'test2', '-n', 'not-default'])

        self.assertEqual(args.name, 'not-default-sebs')

    def test_verbose_level(self):
        args = parse_args(['-b', 'test1', '-b', 'test2', '-vvv'])

        self.assertEqual(args.verbose, 4)


if __name__ == '__main__':
    unittest.main()



import sys
import boto3
import logging
import unittest
import datetime
import botocore.session
from botocore.stub import Stubber, ANY
from unittest.mock import patch, MagicMock, Mock

if('sebs.ec2' in sys.modules):
    # We need to un-import it if imported already
    del sys.modules["sebs.ec2"]


class TestStatefulVolume(unittest.TestCase):

    def setUp(self):
        # Disable logging for testing
        logging.disable(logging.CRITICAL)

        self.instance_id = 'i-1234567890abcdef0'
        self.tag_name = 'sebs'
        self.device_name = '/dev/xdf'

        # Setup our ec2 client stubb
        ec2 = botocore.session.get_session().create_client('ec2', region_name='us-west-2')
        self.ec2_client = ec2
        self.stub_client = Stubber(ec2)

        # Setup our ec2 resource stub
        ec2_resource = boto3.resource('ec2', region_name='us-west-2')
        self.stub_resource = Stubber(ec2_resource.meta.client)

        # Use mocks to pass out client stubb to our code
        self.boto3 = MagicMock(name='module_mock')
        self.mock_client = MagicMock(name='client_mock', return_value=ec2)
        self.boto3.client = self.mock_client

        # Setup resource and volume mocks
        self.mock_snapshot = MagicMock(
            name='snapshot_mock', snapshot_id='sn-12345')
        self.first_volume = Mock(name='first_volume',
                                 attachments=[{'InstanceId': ''}], volume_type='GP2', volume_id='vol-1111')

        self.second_volume = Mock(name='second_volume',
                                  attachments=[{'InstanceId': ''}], volume_type='GP2', volume_id='vol-2222')

        self.first_volume.create_snapshot = Mock(
            return_value=self.mock_snapshot)

        self.mock_resource = MagicMock(
            name='mock_resource_object')

        self.mock_resource.Volume = MagicMock(
            name='mock_volume_constructor')

        self.mock_resource.Volume.side_effect = [
            self.first_volume, self.second_volume]

        self.boto3.resource = MagicMock(
            name='mock_resource_contructor', return_value=self.mock_resource)

        self.mock_session = MagicMock(name='mock_session')

        self.mock_session.client.return_value = self.ec2_client

        self.mock_session.resource.return_value = self.mock_resource

        self.boto3.session.Session = self.mock_session

        modules = {
            'boto3': self.boto3,
            'ec2_metadata': MagicMock()
        }

        self.default_response = {
            'Volumes': [
                {
                    'Attachments': [],
                    'AvailabilityZone': 'string',
                    'CreateTime': datetime.datetime(2015, 1, 1),
                    'Encrypted': False,
                    'KmsKeyId': 'string',
                    'OutpostArn': 'string',
                    'Size': 123,
                    'SnapshotId': 'string',
                    'State': 'available',
                    'VolumeId': 'vol-XXXXXX',
                    'Iops': 123,
                    'Tags': [
                        {
                            'Key': 'string',
                            'Value': 'string'
                        },
                    ],
                    'VolumeType': 'gp2',
                    'FastRestored': False,
                    'MultiAttachEnabled': False
                }
            ]
        }

        self.default_params = {'Filters': ANY}

        self.stub_client.activate()

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()

        from sebs.ec2 import StatefulVolume

        self.StatefulVolume = StatefulVolume

    def tearDown(self):
        # Turn logging back on
        logging.disable(logging.NOTSET)

        self.module_patcher.stop()
        self.stub_client.deactivate()

    def test_class_properties(self):

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        self.stub_client.assert_no_pending_responses()

        self.assertEqual(sv.instance_id, self.instance_id,
                         'Should set the instance_id we pass in.')
        self.assertEqual(sv.device_name, self.device_name,
                         'Should set the deviceName')
        self.assertEqual(
            sv.ready, False, 'Should set the volume to not ready.')
        self.assertEqual(sv.status, 'Unknown', 'Should set the status')
        self.assertEqual(sv.volume, None, 'Should set the tag name')
        self.assertEqual(sv.tag_name, self.tag_name, 'Should set the tag name')
        self.assertEqual(sv.ec2_client, self.ec2_client,
                         'Should set an ec2 client')
        self.assertEqual(sv.ec2_resource, self.mock_resource,
                         'Should set our mock volume.')

    def test_status_new(self):

        response = self.default_response.copy()
        response['Volumes'] = []

        self.stub_client.add_response(
            'describe_volumes', response, self.default_params)

        self.stub_client.add_response(
            'describe_volumes', self.default_response, {'Filters': [
                {
                    'Name': 'attachment.instance-id',
                    'Values': [self.instance_id]
                },
                {
                    'Name': 'attachment.device',
                    'Values': [
                        self.device_name,
                    ]
                }
            ]})

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        sv.get_status()

        self.stub_client.assert_no_pending_responses()
        self.assertEqual(sv.status, 'Attached',
                         'Volume should be mounted already.')
        self.assertIsInstance(sv.volume, Mock,
                              'Should have a boto3 Volume resource')
        self.assertEqual(sv.ready, True, 'Should be ready.')

    def test_status_missing(self):

        response = self.default_response.copy()
        response['Volumes'] = []

        self.stub_client.add_response(
            'describe_volumes', response, self.default_params)

        self.stub_client.add_response(
            'describe_volumes', response, {'Filters': [
                {
                    'Name': 'attachment.instance-id',
                    'Values': [self.instance_id]
                },
                {
                    'Name': 'attachment.device',
                    'Values': [
                        self.device_name,
                    ]
                }
            ]})

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        sv.get_status()

        self.stub_client.assert_no_pending_responses()
        self.assertEqual(sv.status, 'Missing',
                         'We should not find a tagged volume.')
        self.assertEqual(sv.volume, None,
                         'Should not have set a volume resourse.')
        self.assertEqual(sv.ready, False, 'Should not be ready.')

    def test_status_duplicate(self):

        response = self.default_response.copy()
        response['Volumes'] = [
            {'VolumeId': 'vol-1111'}, {'VolumeId': 'vol-2222'}]

        self.stub_client.add_response(
            'describe_volumes', response, self.default_params)

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        sv.get_status()

        self.stub_client.assert_no_pending_responses()
        self.assertEqual(sv.status, 'Duplicate',
                         'Should be a duplicate volume')
        self.assertEqual(sv.volume, None,
                         'Should not have set a volume resourse.')
        self.assertEqual(sv.ready, False, 'Volume should not be Ready')

    def test_status_not_attached(self):

        self.stub_client.add_response(
            'describe_volumes', self.default_response, self.default_params)

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        sv.get_status()

        self.stub_client.assert_no_pending_responses()
        self.assertEqual(sv.status, 'Not Attached',
                         'Should find an existing volume')
        self.assertEqual(sv.volume, self.first_volume,
                         'Should be our volume mock')
        self.assertEqual(sv.ready, False, 'Volume should not be Ready')

    def test_volume_tagging(self):
        response = self.default_response.copy()
        response['Volumes'] = []

        self.stub_client.add_response(
            'describe_volumes', response, {'Filters': [
                {
                    'Name': f'tag:{self.tag_name}',
                    'Values': [self.device_name]
                }
            ]})

        self.stub_client.add_response(
            'describe_volumes', self.default_response, {'Filters': [
                {
                    'Name': 'attachment.instance-id',
                    'Values': [self.instance_id]
                },
                {
                    'Name': 'attachment.device',
                    'Values': [
                        self.device_name,
                    ]
                }
            ]})

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        sv.get_status()

        sv.tag_volume()

        self.first_volume.create_tags.assert_called_with(
            Tags=[{'Key': self.tag_name, 'Value': self.device_name}])

    def test_copy_new(self):
        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)
        sv.status = 'New'

        response = sv.copy('fakeAZ')

        self.assertEqual(response, 'New',
                         "Should do nothing if status in not 'Not Attached'.")

    def test_copy_same_az(self):
        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        self.first_volume.availability_zone = 'fakeAZ'
        sv.status = 'Not Attached'
        sv.volume = self.first_volume

        response = sv.copy('fakeAZ')

        self.assertEqual(response, 'Not Attached',
                         "Should not change the status if in the same AZ.")
        self.first_volume.copy.assert_not_called()

    def test_copy_different_az(self):

        self.stub_client.add_response(
            'create_volume', {'VolumeId': 'vol-2222',
                              'AvailabilityZone': 'newAZ',
                              'Encrypted': True,
                              'Size': 50,
                              'SnapshotId': 'sn-2323',
                              'VolumeType': 'gp2'}, {'AvailabilityZone': 'newAZ',
                                                     'SnapshotId': ANY,
                                                     'VolumeType': ANY,
                                                     'TagSpecifications': ANY})

        self.stub_client.add_response('describe_volumes', {'Volumes': [
                                      {'VolumeId': 'vol-2222', 'State': 'available'}]}, {'VolumeIds': ['vol-2222']})

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        self.first_volume.availability_zone = 'fakeAZ'
        sv.status = 'Not Attached'
        sv.volume = self.mock_resource.Volume()

        response = sv.copy('newAZ')

        # Last test we need here is that we are actually creating a second volume
        self.assertEqual(response, 'Not Attached',
                         'Should not change the status after a copy')
        self.assertFalse(sv.ready, 'We should not be ready after copying.')
        self.assertEqual(sv.volume, self.second_volume,
                         'Should have our second volume.')
        self.first_volume.delete.assert_called_once()
        self.mock_snapshot.delete.assert_called_once()

    def test_attach_new(self):
        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)
        sv.status = 'New'

        response = sv.attach()

        self.assertEqual(response, 'New',
                         "Should do nothing if status in not 'Not Attached'.")

    def test_attach(self):

        self.stub_client.add_response('describe_volumes', {'Volumes': [
                                      {'VolumeId': 'vol-1111', 'State': 'in-use'}]}, {'Filters': [
                                          {
                                              'Name': 'attachment.instance-id',
                                              'Values': [
                                                  self.instance_id,
                                              ]
                                          },
                                          {
                                              'Name': 'attachment.device',
                                              'Values': [
                                                  self.device_name,
                                              ]
                                          },
                                      ]})

        self.stub_client.add_response('describe_volumes', {'Volumes': [
                                      {'VolumeId': 'vol-2222', 'State': 'available'}]}, {'VolumeIds': ['vol-1111']})

        self.stub_client.add_response('describe_volumes', {'Volumes': [
                                      {'VolumeId': 'vol-2222', 'State': 'in-use'}]}, {'VolumeIds': ['vol-2222']})

        sv = self.StatefulVolume(self.mock_session,
                                 self.instance_id,
                                 self.device_name,
                                 self.tag_name)

        sv.status = 'Not Attached'
        sv.volume = self.second_volume

        response = sv.attach()

        self.assertEqual(response, 'Attached',
                         'Should be Attached to the instance.')
        self.assertFalse(sv.ready, 'Should not be Ready')
        # We shoud delete the previous volume
        self.first_volume.delete.assert_called_once()


if __name__ == '__main__':
    unittest.main()



import unittest
from sebs.ec2 import Instance
from unittest.mock import patch, MagicMock, call, PropertyMock


class TestInstance(unittest.TestCase):

    def setUp(self):
        self.default_tag = 'sebs'
        self.device_name = '/dev/xdf'
        self.mock_instance = MagicMock(name='mock_instance', id='in-1111')

    def tearDown(self):
        pass

    @patch('sebs.ec2.Instance.get_instance')
    def test_class_properites(self, mock_method):
        mock_method.return_value = self.mock_instance

        server = Instance(self.default_tag)

        self.assertEqual(server.instance, self.mock_instance,
                         'Should set an instance field.')
        self.assertEqual(server.volume_tag, self.default_tag,
                         'Should set volume_tag field.')
        self.assertEqual(server.backup, [], 'Should have empty backup list.')

        mock_method.assert_called_once()

    @patch('sebs.ec2.StatefulVolume')
    @patch('sebs.ec2.Instance.get_instance')
    def test_add_device(self, mock_method, mock_volume_class):
        mock_method.return_value = self.mock_instance
        mock_volume = MagicMock(name='mock_volume')
        mock_volume_class.return_value = mock_volume

        server = Instance(self.default_tag)

        server.add_stateful_device(self.device_name)

        mock_method.assert_called_once()
        mock_volume_class.assert_called_once()

        # Instead of None it should be a session...
        # but I can't figure out how to make a mock method (get_instance)
        # set a class field self.session as a side effect.
        mock_volume_class.assert_called_once_with(None,
                                                  self.mock_instance.id,
                                                  self.device_name,
                                                  self.default_tag)
        self.assertIn(mock_volume, server.backup,
                      'Should put our device in the backup list.')
        mock_volume.get_status.assert_called_once()

    @patch('sebs.ec2.StatefulVolume')
    @patch('sebs.ec2.Instance.get_instance')
    def test_add_multiple_devices(self, mock_method, mock_volume_class):
        mock_method.return_value = self.mock_instance
        mock_volume = MagicMock(name='mock_volume_1')
        mock_volume_class.return_value = mock_volume

        server = Instance(self.default_tag)

        server.add_stateful_device(self.device_name)
        server.add_stateful_device('/dev/2')

        mock_method.assert_called_once()
        self.assertEqual(mock_volume_class.call_count, 2,
                         'Should create two volumes.')

        # Same as above test, need to replace None with session
        mock_volume_class.assert_has_calls(
            [call(None, self.mock_instance.id, self.device_name, self.default_tag),
             call(None, self.mock_instance.id, '/dev/2', self.default_tag)])

        self.assertEqual(len(server.backup), 2, 'Should have two volumes.')
        self.assertEqual(mock_volume.get_status.call_count, 2,
                         'Should have called get_status twice')

    @patch('sebs.ec2.Instance.get_instance')
    def test_tag_volumes(self, mock_method):
        server = Instance(self.default_tag)
        mock_volume = MagicMock(name='mock_volume_1', status='Not Attached')
        mock_volume2 = MagicMock(name='mock_volume_2', status='Missing')

        server = Instance(self.default_tag)
        server.backup = [mock_volume, mock_volume2]
        server.tag_stateful_volumes()

        # Should tag one volume but not the other.
        mock_volume.tag_volume.assert_called_once()
        mock_volume2.tag_volume.assert_not_called()

    @patch('sebs.ec2.ec2_metadata')
    @patch('sebs.ec2.Instance.get_instance')
    def test_attach_volumes(self, mock_method, mock_metadata):

        p = PropertyMock(return_value='AZ2')
        type(mock_metadata).availability_zone = p

        mock_volume = MagicMock(name='mock_volume_1', status='Not Attached')
        mock_volume2 = MagicMock(name='mock_volume_2', status='Missing')

        server = Instance(self.default_tag)
        server.backup = [mock_volume, mock_volume2]
        server.attach_stateful_volumes()

        # Should Attach one volume but not the other.
        mock_volume.copy.assert_called_once_with('AZ2')
        mock_volume.attach.assert_called_once()

        mock_volume2.copy.assert_not_called()
        mock_volume2.attach.assert_not_called()


if __name__ == '__main__':
    unittest.main()



import time
import json
import boto3
import unittest
import warnings
from tests.utils import aws_utils
from botocore.exceptions import ClientError


class TestSebs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print('Setting up environment...')

        warnings.filterwarnings(
            "ignore", category=ResourceWarning, message="unclosed.*<ssl.SSLSocket.*>")

        iam_resources = aws_utils.create_iam_resources()

        cls.iam_role = iam_resources['role']

        cls.iam_role_policy = iam_resources['policy']

        cls.instance_profile = iam_resources['profile']

        cls.default_user_data = aws_utils.create_default_userdata()

        ami_id = aws_utils.get_latest_ami()

        cls.default_instance = dict(BlockDeviceMappings=[],
                                    ImageId=ami_id,
                                    InstanceType='t1.micro',
                                    MaxCount=1,
                                    MinCount=1,
                                    Monitoring={'Enabled': False},
                                    UserData=cls.default_user_data,
                                    IamInstanceProfile={
            'Arn': cls.instance_profile.arn},
            TagSpecifications=[
            {
                'ResourceType': 'instance',
                'Tags': [
                    {
                        'Key': 'Name',
                        'Value': 'Test'
                    },
                ]
            }],
        )

        print('Environment setup.')

    @classmethod
    def tearDownClass(cls):
        print('Cleaning up environment...')

        if cls.instance_profile:
            print('Removing role from instance profile.')
            cls.instance_profile.remove_role(
                RoleName=cls.iam_role.name
            )

        if cls.iam_role_policy:
            print(f'Deleting policy: {cls.iam_role_policy}')
            cls.iam_role_policy.delete()

        if cls.iam_role:
            print(f'Deleting Role: {cls.iam_role}')
            cls.iam_role.delete()

        if cls.instance_profile:
            print(f'Deleting Instance Profile: {cls.instance_profile}')
            cls.instance_profile.delete()

        print('Environment Cleanup Finished')

    def setUp(self):

        print('Setting up test.')
        self.volume_cleanup = []
        self.instance_cleanup = []

        self.ec2 = boto3.resource('ec2')

        self.iam = boto3.resource('iam')

        print('Finished test setup.')

    def tearDown(self):

        print('Running Test Cleanup...')
        for instance in self.instance_cleanup:
            print(f'Deleting {instance.id}')
            instance.terminate()

            waiter = aws_utils.get_ec2_waiter('instance_terminated')

            waiter.wait(InstanceIds=[instance.id])

        for volume in self.volume_cleanup:
            print(f'Deleting {volume.id}')
            try:
                volume.delete()
            except ClientError as e:
                if e.response['Error']['Code'] != 'InvalidVolume.NotFound':
                    raise e

        print('Test Cleanup Finished')

    def test_new_volume(self):

        print('Starting test_new_volume')
        control_tag = 'new-volume-sebs'
        device_name = '/dev/xvdh'

        server_config = self.__class__.default_instance.copy()
        server_config['BlockDeviceMappings'].append(
            aws_utils.create_block_device(device_name))

        server_config['UserData'] += (
            f'/usr/local/bin/sebs -b {device_name} -n {control_tag} -vvv\n'
            f'while [ ! -e {device_name} ] ; do sleep 1 ; done\n'
        )

        instance = aws_utils.create_instance(server_config)

        self.instance_cleanup.append(instance)

        instance.wait_until_running()
        instance.reload()

        volume_id = aws_utils.get_volume_from_bdm(device_name, instance)

        self.assertTrue(volume_id, 'Should have the new volume attached')

        waiter = aws_utils.get_ec2_waiter('volume_in_use')
        waiter.wait(VolumeIds=[volume_id])

        volume = self.ec2.Volume(volume_id)

        volume_tagged = aws_utils.has_control_tag(
            control_tag, device_name, volume)

        self.assertTrue(volume_tagged, 'Volume should have our control tag.')

        print('Finshed test_new_volume')

    def test_new_volumes(self):

        print('Starting test_new_volumes')

        control_tag = 'new-volumes-sebs'
        device1_name = '/dev/xvdh'
        device2_name = '/dev/xvdm'

        server_config = self.__class__.default_instance.copy()
        server_config['BlockDeviceMappings'].append(
            aws_utils.create_block_device(device1_name))
        server_config['BlockDeviceMappings'].append(
            aws_utils.create_block_device(device2_name))

        server_config['UserData'] += (
            f'/usr/local/bin/sebs -b {device1_name} -b {device2_name} -n {control_tag} -vvv\n'
            f'while [ ! -e {device1_name} ] ; do sleep 1 ; done\n'
            f'while [ ! -e {device2_name} ] ; do sleep 1 ; done\n'
        )

        instance = aws_utils.create_instance(server_config)

        self.instance_cleanup.append(instance)

        instance.wait_until_running()
        instance.reload()

        volume1_id = aws_utils.get_volume_from_bdm(device1_name, instance)
        volume2_id = aws_utils.get_volume_from_bdm(device2_name, instance)

        self.assertTrue(
            volume1_id, 'Should have the new volume attached')
        self.assertTrue(
            volume2_id, 'Should have the new volume attached')

        waiter = aws_utils.get_ec2_waiter('volume_in_use')
        waiter.wait(VolumeIds=[volume1_id])
        waiter.wait(VolumeIds=[volume2_id])

        volume1 = self.ec2.Volume(volume1_id)
        volume2 = self.ec2.Volume(volume2_id)

        volume1_tagged = aws_utils.has_control_tag(
            control_tag, device1_name, volume1)
        volume2_tagged = aws_utils.has_control_tag(
            control_tag, device2_name, volume2)

        self.assertTrue(volume1_tagged, 'Volume should have our control tag.')
        self.assertTrue(volume2_tagged, 'Volume should have our control tag.')

        print('Finished test_new_volumes')

    def test_existing_volume(self):

        print('Starting test_existing_volume')

        control_tag = 'existing-volume-sebs'
        device_name = '/dev/xvdh'

        az = aws_utils.get_avaliable_az()

        existing_vol = aws_utils.create_existing_volume(
            control_tag, device_name, az[0])

        self.volume_cleanup.append(existing_vol)

        server_config = self.__class__.default_instance.copy()
        server_config['BlockDeviceMappings'].append(
            aws_utils.create_block_device(device_name))
        server_config['Placement'] = {
            'AvailabilityZone': az[1]
        }

        server_config['UserData'] += (
            f'/usr/local/bin/sebs -b {device_name} -n {control_tag} -vvv\n'
            f'while [ ! -e {device_name} ] ; do sleep 1 ; done\n'
        )

        instance = aws_utils.create_instance(server_config)

        self.instance_cleanup.append(instance)

        instance.wait_until_running()
        instance.reload()

        volume_id = aws_utils.get_volume_from_bdm(device_name, instance)

        default_vol = self.ec2.Volume(volume_id)

        self.assertTrue(volume_id, 'Should have the default volume attached')

        waiter = aws_utils.get_ec2_waiter('volume_deleted')
        waiter.wait(VolumeIds=[volume_id])

        instance.reload()

        new_vol_id = aws_utils.get_volume_from_bdm(device_name, instance)

        new_vol = self.ec2.Volume(new_vol_id)
        self.volume_cleanup.append(new_vol)

        self.assertNotEqual(new_vol.id, default_vol.id,
                            'Volume attached now should not be what we started with.')

        volume_tagged = aws_utils.has_control_tag(
            control_tag, device_name, new_vol)

        waiter = aws_utils.get_ec2_waiter('volume_deleted')

        waiter.wait(VolumeIds=[existing_vol.id, default_vol.id])

        self.assertTrue(volume_tagged, 'Volume should have our control tag.')

        print('Finished test_existing_volume')

    def test_existing_volumes(self):

        print('Starting test_existing_volumes')

        control_tag = 'existing-volumes-sebs'
        device1_name = '/dev/xvdh'
        device2_name = '/dev/xvdm'

        az = aws_utils.get_avaliable_az()

        existing_vol1 = aws_utils.create_existing_volume(
            control_tag, device1_name, az[0])

        existing_vol2 = aws_utils.create_existing_volume(
            control_tag, device2_name, az[0])

        self.volume_cleanup.extend([existing_vol1, existing_vol2])

        server_config = self.__class__.default_instance.copy()
        server_config['BlockDeviceMappings'].append(
            aws_utils.create_block_device(device1_name))
        server_config['BlockDeviceMappings'].append(
            aws_utils.create_block_device(device2_name))
        server_config['Placement'] = {
            'AvailabilityZone': az[1]
        }

        server_config['UserData'] += (
            f'/usr/local/bin/sebs -b {device1_name} -b {device2_name} -n {control_tag} -vvv\n'
            f'while [ ! -e {device1_name} ] ; do sleep 1 ; done\n'
            f'while [ ! -e {device2_name} ] ; do sleep 1 ; done\n'
        )

        instance = aws_utils.create_instance(server_config)

        self.instance_cleanup.append(instance)

        instance.wait_until_running()
        instance.reload()

        default1_vol_id = aws_utils.get_volume_from_bdm(device1_name, instance)
        default2_vol_id = aws_utils.get_volume_from_bdm(device2_name, instance)

        default_vol1 = self.ec2.Volume(default1_vol_id)
        default_vol2 = self.ec2.Volume(default2_vol_id)

        self.volume_cleanup.extend([default_vol1, default_vol2])

        self.assertTrue(default1_vol_id,
                        'Should have the default volume attached')
        self.assertTrue(default2_vol_id,
                        'Should have the default volume attached')

        waiter = aws_utils.get_ec2_waiter('volume_deleted')
        waiter.wait(VolumeIds=[default1_vol_id])
        waiter.wait(VolumeIds=[default2_vol_id])

        instance.reload()

        new1_vol_id = aws_utils.get_volume_from_bdm(device1_name, instance)
        new2_vol_id = aws_utils.get_volume_from_bdm(device2_name, instance)

        new_vol1 = self.ec2.Volume(new1_vol_id)
        new_vol2 = self.ec2.Volume(new2_vol_id)

        self.volume_cleanup.extend([new_vol1, new_vol2])

        self.assertNotEqual(new_vol1.id, default_vol1.id,
                            'Volume attached now should not be what we started with.')
        self.assertNotEqual(new_vol2.id, default_vol2.id,
                            'Volume attached now should not be what we started with.')

        volume1_tagged = aws_utils.has_control_tag(
            control_tag, device1_name, new_vol1)

        volume2_tagged = aws_utils.has_control_tag(
            control_tag, device2_name, new_vol2)

        waiter = aws_utils.get_ec2_waiter('volume_deleted')

        waiter.wait(VolumeIds=[
                    existing_vol1.id,
                    default_vol1.id,
                    existing_vol2.id,
                    default_vol2.id
                    ])

        self.assertTrue(volume1_tagged, 'Volume should have our control tag.')
        self.assertTrue(volume2_tagged, 'Volume should have our control tag.')

        print('Finished test_existing_volumes')



import unittest
import argparse
from sebs.app import main
from unittest.mock import MagicMock, patch


class TestApplicaton(unittest.TestCase):

    @patch('sebs.app.Instance')
    @patch('sebs.ec2.ec2_metadata')
    def test_main(self, mock_metadata, mock_class):
        mock_instance = MagicMock(name='mock_instance')
        mock_class.return_value = mock_instance

        args = argparse.Namespace(name='sebs', backup=['/dev/xdv', '/dev/svh'])
        with self.assertRaises(SystemExit):
            main(args)

        mock_class.assert_called_once_with('sebs')
        mock_instance.add_stateful_device.assert_any_call('/dev/xdv')
        mock_instance.add_stateful_device.assert_any_call('/dev/svh')
        mock_instance.attach_stateful_volumes.assert_called_once()
        mock_instance.tag_stateful_volumes.assert_called_once()


if __name__ == '__main__':
    unittest.main()



import os
import datetime
import subprocess
import setuptools


release_version = os.getenv('RELEASE_VERSION')

if not release_version:
    tag = subprocess.check_output(
        ["git", "describe", "--abbrev=0"]
    ).strip().decode('ASCII')

    now = datetime.datetime.now()

    release_version = f'{tag}.dev{now.hour}{now.minute}'


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sebs",
    version=release_version,
    author="Levi Blaney",
    author_email="shadycuz@gmail.com",
    description="Create Stateful Elastic Block Storage on AWS.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DontShaveTheYak/sebs",
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=[
        'boto3',
        'ec2-metadata',
        'importlib-metadata ~= 1.0 ; python_version < "3.8"'
    ],
    scripts=['bin/sebs'],
    python_requires='>=3.6',
)



import os
import time
import json
import boto3
import subprocess


def create_iam_resources():
    print('Creating IAM resources for testing.')
    iam = boto3.resource('iam')

    assume_role_policy_doc = {'Version': '2012-10-17'}

    assume_role_policy_doc['Statement'] = [{
        'Action': [
            "sts:AssumeRole"
        ],
        'Effect': 'Allow',
        'Principal': {
            'Service': ["ec2.amazonaws.com"]
        }
    }]

    iam_role = iam.create_role(
        RoleName='Sebs',
        AssumeRolePolicyDocument=json.dumps(
            assume_role_policy_doc, indent=2),
        Description='Used for functional testing.',
    )

    role_policy_doc = {'Version': '2012-10-17'}

    role_policy_doc['Statement'] = [
        {
            'Action': [
                "ec2:DetachVolume",
                "ec2:AttachVolume",
                "ec2:DeleteVolume",
                "ec2:DeleteSnapshot",
                "ec2:CreateTags",
                "ec2:CreateSnapshot",
                "ec2:CreateVolume"
            ],
            'Effect': 'Allow',
            'Resource': ["arn:aws:ec2:*:*:instance/*",
                         "arn:aws:ec2:*::snapshot/*",
                         "arn:aws:ec2:*:*:volume/*"]
        },
        {
            'Action': ["ec2:DescribeInstances", "ec2:DescribeVolumes", "ec2:DescribeSnapshots"],
            'Effect': 'Allow',
            'Resource': ["*"]
        }
    ]

    iam_role_policy = iam_role.Policy('Create-Volumes')

    iam_role_policy.put(
        PolicyDocument=json.dumps(role_policy_doc, indent=2)
    )

    instance_profile = iam.create_instance_profile(
        InstanceProfileName='Sebs-EC2-Profile',
    )

    instance_profile.add_role(
        RoleName=iam_role.name
    )
    print('Waiting on instance profile creation.')
    waiter = boto3.client('iam').get_waiter('instance_profile_exists')

    waiter.wait(
        InstanceProfileName=instance_profile.name
    )

    print('IAM resources created.')

    return {
        'role': iam_role,
        'policy': iam_role_policy,
        'profile': instance_profile
    }


def create_default_userdata():

    git_ref = os.getenv('GITHUB_SOURCE_BRANCH')

    if not git_ref:
        git_ref = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"]
        ).strip().decode('ASCII')

    return (
        "#!/bin/bash\n"
        "exec > >(tee /var/log/user-data.log|logger -t user-data -s 2>/dev/console) 2>&1\n"
        "yum install python3 git -y\n"
        f"python3 -m pip install git+https://github.com/DontShaveTheYak/sebs.git@{git_ref}#egg=sebs --upgrade \n"
        "export AWS_METADATA_SERVICE_NUM_ATTEMPTS=3 \n"
        "export AWS_METADATA_SERVICE_TIMEOUT=2 \n"
    )


def get_latest_ami():
    ec2 = boto3.resource('ec2')
    images = ec2.images.filter(Owners=['amazon'],
                               Filters=[
                                   {'Name': 'name',
                                    'Values': ['amzn2*']
                                    }])

    for image in images:
        ami = image
        break

    return ami.id


def create_block_device(device_name):
    return {
        'DeviceName': device_name,
        'Ebs': {
            'DeleteOnTermination': True,
            'VolumeSize': 10,
            'VolumeType': 'gp2',
            'Encrypted': False
        }
    }


def create_instance(instance_config):
    ec2 = boto3.resource('ec2')
    server_list = ec2.create_instances(**instance_config)

    instance = server_list[0]

    return instance


def get_volume_from_bdm(device_name, instance):
    volume_id = ''
    for device in instance.block_device_mappings:
        if device['DeviceName'] == device_name:
            volume_id = device['Ebs']['VolumeId']
            break

    return volume_id


def get_ec2_waiter(name):
    waiter = boto3.client('ec2').get_waiter(name)
    return waiter


def has_control_tag(control_tag, device_name, volume):

    wait_for_volume_tag(volume)

    tags = [tag for tag in volume.tags if tag['Key']
            == control_tag and tag['Value'] == device_name]

    return bool(tags)


def wait_for_volume_tag(volume):
    i = 0
    tagged = False
    while True:
        i += 1

        volume.reload()

        if volume.tags:
            tagged = True
            break

        time.sleep(30)

        if i > 14:
            break

    if not tagged:
        raise Exception(f'{volume.id} was never tagged.')


def get_default_vpc():
    ec2 = boto3.resource('ec2')

    vpcs = ec2.vpcs.all()

    default_vpc = next(vpc for vpc in vpcs if vpc.is_default)

    return default_vpc


def get_avaliable_az():

    vpc = get_default_vpc()

    az = [subnet.availability_zone for subnet in vpc.subnets.all()]

    return az


def create_existing_volume(control_tag, device_name, az):
    ec2 = boto3.resource('ec2')

    volume = ec2.create_volume(
        AvailabilityZone=az,
        Encrypted=False,
        Size=10,
        VolumeType='gp2',
        TagSpecifications=[
            {
                'ResourceType': 'volume',
                'Tags': [
                    {
                        'Key': control_tag,
                        'Value': device_name
                    }
                ]
            }
        ]
    )

    waiter = get_ec2_waiter('volume_available')

    waiter.wait(VolumeIds=[volume.id])

    return volume
